#!python3
# -*- coding: utf-8 -*-
"""
Data model to represent a CKAN database architecture
"""

ckan_tags_sep = ','   # separator for tags, fields in primary key, indexes, queries
ckan_timestamp_sep = 'T'  # separator for timestamp ISO format
environ_keyword = "environ"

